import debounce from 'lodash.debounce'
import {
  CharacteristicEventTypes,
  CharacteristicGetCallback,
  CharacteristicSetCallback,
  CharacteristicValue,
  Service,
} from 'homebridge'
import { SharedFoundation } from './shared/foundation'
import { SharedDevice } from './shared/device'
import {
  DeviceInstance,
  MIIOSpecMapping,
  MIIOSpecResponseValueMapping,
  MIoTDeviceIdentify,
  MIoTSpecMapping, MIoTSpecsGetQuery,
  MIoTSpecsResponseValueMapping, MIIORegisterConfig, MIoTRegisterConfig, PropertyValue
} from './main'

type Props = {
  identify: MIoTDeviceIdentify
  characteristicsService: Service
}

enum ErrorMessages {
  NotConnect = 'Device not connected.',
  SpecNotFound = 'Spec not found.',
}

export class MIoTDevice {

  // Requirement
  private readonly characteristicsService: Service
  // Device
  private readonly identify: MIoTDeviceIdentify
  // Properties: MIoT
  private MIoTPreviousProperties: MIoTSpecsResponseValueMapping = {}
  private MIoTSpecsMapping: MIoTSpecMapping = {}
  private MIoTSpecsQueue: ((valueMapping: MIoTSpecsResponseValueMapping) => void)[] = []
  // Properties: MIIO
  private MIIOPreviousProperties: MIIOSpecResponseValueMapping = {}
  private MIIOSpecsMapping: MIIOSpecMapping = {}
  private MIIOSpecsQueue: ((valueMapping: MIIOSpecResponseValueMapping) => void)[] = []

  constructor (props: Props) {
    // HomeBridge
    this.characteristicsService = props.characteristicsService
    // Device
    this.identify = props.identify
  }

  /*
   * MIoT
   */
  // Spec
  public addMIoTSpec = (specs: MIoTSpecMapping) => {
    Object.values(specs).forEach(i => this.MIoTSpecsMapping[i.name] = i)
  }
  private getMIoTSpec = async (device: DeviceInstance, name?: string | string[]) => {
    // Action: getAll
    if (!name) return Object.values(this.MIoTSpecsMapping).map(i => ({ ...i, did: device.did }))
    // Action: getByName
    const targetSpecs: MIoTSpecsGetQuery = []
    if (Array.isArray(name)) {
      name.forEach(i => {
        const spec = this.MIoTSpecsMapping[i]
        if (spec) targetSpecs.push({ ...spec, did: device.did })
      })
    } else {
      const spec = this.MIoTSpecsMapping[name]
      if (spec) targetSpecs.push({ ...spec, did: device.did })
    }
    return targetSpecs
  }
  // Properties
  private debounceRequestMIoTProperty = debounce(async () => {
    // Guard
    const device = await SharedDevice.getInstance(this.identify)
    if (!device?.did) throw new Error(ErrorMessages.NotConnect)
    // Spec
    const targetSpecs = await this.getMIoTSpec(device)
    // Pull queue
    const queue = [...this.MIoTSpecsQueue]
    this.MIoTSpecsQueue = []
    // Get properties
    const properties = await device.miioCall('get_properties', targetSpecs)
    this.MIoTPreviousProperties = targetSpecs.reduce((acc, cur, idx) => ({
      ...acc,
      [cur.name]: properties[idx].value
    }), {} as MIoTSpecsResponseValueMapping)
    SharedFoundation.log.debug(`MIoT Merging request of ${this.identify.name} ${this.identify.address}`)
    queue.forEach(resolve => resolve(this.MIoTPreviousProperties))
  }, SharedFoundation.config.requestPropertyDebounceThreshold)
  private setMIoTProperty = async (name: string, value: PropertyValue) => {
    // Guard
    const device = await SharedDevice.getInstance(this.identify)
    if (!device?.did) throw new Error(ErrorMessages.NotConnect)
    // Spec
    const targetSpec = this.MIoTSpecsMapping[name]
    if (!targetSpec) throw new Error(ErrorMessages.SpecNotFound)
    // Action
    return device.miioCall('set_properties', [Object.assign(targetSpec, { value, did: device.did })])
  }
  // Events
  private pullMIoTProperty = async (): Promise<MIoTSpecsResponseValueMapping> => {
    // Action
    return new Promise((resolve => {
      // Queue update
      this.MIoTSpecsQueue.push(resolve)
      // Trigger Property getter
      this.debounceRequestMIoTProperty()
    }))
  }
  public addMIoTCharacteristicListener(type: any, config: MIoTRegisterConfig) {
    const characteristic = this.characteristicsService.getCharacteristic(type)
    if ('get' in config) {
      characteristic.on(CharacteristicEventTypes.GET, async (callback: CharacteristicGetCallback) => {
        try {
          SharedFoundation.log.debug(`MIoT START GETTING ${type.name}`, Date.now())
          const property = await this.pullMIoTProperty()
          const propertyFormatted = config.get.formatter(property)
          callback(undefined, propertyFormatted)
          SharedFoundation.log.debug(`MIoT GETTING ${type.name} SUCCESS `, propertyFormatted)
        } catch (e) {
          SharedFoundation.log.error(`MIoT GETTING ${type.name} ERROR`, e)
          callback(e)
        }
      })
    }
    if ('set' in config) {
      const set = config.set
      if (set) {
        characteristic.on(CharacteristicEventTypes.SET, async (value: CharacteristicValue, callback: CharacteristicSetCallback) => {
          try {
            SharedFoundation.log.debug(`MIoT START SETTING ${type.name}`, Date.now())
            const valueFormatted = set.formatter(value, this.MIoTPreviousProperties)
            await this.setMIoTProperty(set.property, valueFormatted)
            callback(undefined, value)
            SharedFoundation.log.debug(`MIoT SETTING ${type.name} SUCCESS`, valueFormatted)
          } catch (e) {
            SharedFoundation.log.error(`MIoT SETTING ERROR ${type.name}`, e)
            callback(e)
          }
        })
      }
    }
  }

  /*
   * MIIO
   */
  // Spec
  public addMIIOSpec = (specs: MIIOSpecMapping) => {
    Object.values(specs).forEach(i => this.MIIOSpecsMapping[i] = i)
  }
  private getMIIOSpec = async (name?: string | string[]) => {
    // Action: getAll
    if (!name) return Object.values(this.MIIOSpecsMapping)
    // Action: getByName
    if (Array.isArray(name)) {
      return name
    } else {
      return [name]
    }
  }
  // Properties
  private debounceRequestMIIOProperty = debounce(async () => {
    // Guard
    const device = await SharedDevice.getInstance(this.identify)
    if (!device?.did) throw new Error(ErrorMessages.NotConnect)
    // Spec
    const targetSpecs = await this.getMIIOSpec()
    // Pull queue
    const queue = [...this.MIIOSpecsQueue]
    this.MIIOSpecsQueue = []
    // Get properties
    const properties = await device.miioCall('get_prop', targetSpecs)
    SharedFoundation.log.debug('MIIO Properties', properties)
    this.MIIOPreviousProperties = targetSpecs.reduce((acc, cur, idx) => ({
      ...acc,
      [cur]: properties[idx]
    }), {} as MIIOSpecResponseValueMapping)
    SharedFoundation.log.debug(`MIIO Merging request of ${this.identify.name} ${this.identify.address}`)
    queue.forEach(resolve => resolve(this.MIIOPreviousProperties))
  }, SharedFoundation.config.requestPropertyDebounceThreshold)
  private setMIIOProperty = async (name: string, value: PropertyValue[]) => {
    // Guard
    const device = await SharedDevice.getInstance(this.identify)
    if (!device?.did) throw new Error(ErrorMessages.NotConnect)
    // Action
    return device.miioCall(name, value)
  }
  // Events
  private pullMIIOProperty = async (): Promise<MIIOSpecResponseValueMapping> => {
    // Action
    return new Promise((resolve => {
      // Queue update
      this.MIIOSpecsQueue.push(resolve)
      // Trigger Property getter
      this.debounceRequestMIIOProperty()
    }))
  }
  public addMIIOCharacteristicListener(type: any, config: MIIORegisterConfig) {
    const characteristic = this.characteristicsService.getCharacteristic(type)
    if ('get' in config) {
      characteristic.on(CharacteristicEventTypes.GET, async (callback: CharacteristicGetCallback) => {
        try {
          SharedFoundation.log.debug(`MIIO START GETTING ${type.name}`, Date.now())
          const property = await this.pullMIIOProperty()
          const propertyFormatted = config.get.formatter(property)
          callback(undefined, propertyFormatted)
          SharedFoundation.log.debug(`MIIO GETTING ${type.name} SUCCESS`, propertyFormatted)
        } catch (e) {
          SharedFoundation.log.error(`MIIO GETTING ${type.name} ERROR`, e)
          callback(e)
        }
      })
    }
    if ('set' in config) {
      const set = config.set
      if (set) {
        characteristic.on(CharacteristicEventTypes.SET, async (value: CharacteristicValue, callback: CharacteristicSetCallback) => {
          try {
            SharedFoundation.log.debug(`MIIO START SETTING ${type.name}`, Date.now())
            const valueFormatted = set.formatter(value, this.MIIOPreviousProperties)
            await this.setMIIOProperty(set.property, valueFormatted)
            callback(undefined, value)
            SharedFoundation.log.debug(`MIIO SETTING ${type.name} SUCCESS`, valueFormatted)
          } catch (e) {
            SharedFoundation.log.error(`MIIO SETTING ${type.name} ERROR`, e)
            callback(e)
          }
        })
      }
    }
  }
}
