// @ts-ignore
import miio from 'miio'
import debounce from 'lodash.debounce'
import { SharedQueue } from './queue'
import { SharedFoundation } from './foundation'
import { getDeviceId, sleep } from '../MIoTDevice.utils'
import { DeviceInstance, MIoTDeviceIdentify } from '../main'

class Device {

  private device: { [address: string]: DeviceInstance | undefined } = {}
  private pool: { [address: string]: ReturnType<typeof debounce> } = {}

  private isConnected = (identify: MIoTDeviceIdentify) => {
    SharedFoundation.log.debug(`SHARED DEVICE CHECKING CONNECTING ${identify.name} ${identify.address}`, Date.now())
    const deviceInstance = this.device[identify.address]
    if (!deviceInstance) return false
    const now = Date.now()
    const diff = now - deviceInstance.timeout
    deviceInstance.timeout = now
    return diff < SharedFoundation.config.reConnectThreshold
  }
  private requestConnect = async (identify: MIoTDeviceIdentify) => {
    // Do connect
    try {
      // Log
      SharedFoundation.log.info(`${identify.name} ${identify.address} start ${!!this.device[identify.address] ? 're-' : ''}connecting.`)
      // Create device instance
      const device = await miio.device({
        address: identify.address,
        token: identify.token
      })
      device.did = getDeviceId(device.id)
      device.timeout = Date.now()
      // Update
      this.device[identify.address] = device
      // Resolve queues
      SharedQueue.resolve(identify, device)
      // Log
      SharedFoundation.log.info(`${identify.name} ${identify.address} connected.`)
    } catch (e) {
      // Retry if failure
      if (this.isConnected(identify)) return
      SharedFoundation.log.info(`${identify.name} ${identify.address} ${identify.token} connect failure, reconnecting ...`)
      SharedFoundation.log.error(e)
      await sleep(SharedFoundation.config.reConnectFailureThreshold)
      await this.requestConnect(identify)
    }
  }
  private debounceRequestConnect (identify: MIoTDeviceIdentify) {
    if (typeof this.pool[identify.address] !== 'function') {
      this.pool[identify.address] = debounce(() => this.requestConnect(identify), SharedFoundation.config.requestConnectDebounceThreshold)
    }
    this.pool[identify.address]()
  }

  public getInstance(identify: MIoTDeviceIdentify): Promise<DeviceInstance> {
    SharedFoundation.log.debug(`SHARED DEVICE GETTING INSTANCE ${identify.name} ${identify.address}`, Date.now())
    // return this.device[identify]
    return new Promise(resolve => {
      // Queue update
      SharedQueue.append(identify, resolve)
      // Trigger getter
      this.debounceRequestConnect(identify)
    })
  }

}

export const SharedDevice = new Device()

